# -*- coding: utf-8; indent-tabs-mode: nil -*-

# mailbridge: a mail storage spanning bridge
# 
# Copyright 2011 Ryan Hughes <ryan@n-h.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from mailbridge.store.zarafa.ZarafaServer import ZarafaServer
from mailbridge.store.zarafa.ZarafaStore import ZarafaStore
from mailbridge.store.zarafa.ZarafaStoreMonitor import *
from mailbridge.store.dsync.DsyncStoreMonitor import *
from mailbridge.store.dsync.DsyncStore import DsyncStore
import MAPI
from MAPI.Util import *
from MAPI.Time import *
from MAPI.Struct import *
import os
import binascii
import time
import logging
import sys
import locale
import threading
import traceback




class CacheUpdater:
    def __init__(self, sinkmonitor, notifsinkmonitor, maillogmonitor, cache_state, logger, backend_account, cache_account, config):
        self.advisesink = sinkmonitor.GetChangeAdviseSink(self)
        self.notifsinkmonitor = notifsinkmonitor
        self.maillogmonitor = maillogmonitor
        self.cache_state = cache_state
        self.logger = logger
        self.relations = {}
        self.cache_folder_path_folder_id = {}
        self.synclock = threading.Lock()
        
        self.backend_account = backend_account
        self.backend_password = None
        self.cache_host = None
        self.cache_ssl_user = None
        self.cache_ssl_password = None
        self.cache_account = cache_account
        self.config = config
        
        self.maillogmonitor.addUpdater(self.cache_account, self)

    def update_backend_for_folder(self, folderid):
        cache_state = self.cache_state
        backend = self.backend
        exporter_flags = self.exporter_flags
        imapcache = self.imapcache
        imapcache.reopen()
        # get cache importer
        cache_folderid = backend.getCacheFolderId(folderid, imapcache)
        cache_folder_path = imapcache.getFolderPath(cache_folderid)
        if (cache_folder_path == None):
            return None
        exporter = imapcache.getExporter(cache_folderid)
        exporter.session = backend.session
        exporter.store = backend.store
        exporter.addressbook = backend.addressbook
    
        state_stream = cache_state.load_state(binascii.hexlify(folderid))
    
        importer = backend.getContentsImporter(folderid)
        importer.Config(state_stream, 0)
        syncstate=None
        exporter.Config(importer, syncstate, 0)
        
        progress = 0
        i = 0
        while True:
            res = exporter.Synchronize(progress);
            if (res[0] == res[1]):
                break;
            progress = res[1]
            i += 1
        stream = IStream()
        importer.UpdateState(stream)
        cache_state.save_state(binascii.hexlify(folderid), stream)

        self.cache_folder_path_folder_id[cache_folder_path] = folderid
        imapcache.close_connection()
        
    def update_imap_cache_for_folder(self, folderid):
        cache_state = self.cache_state
        backend = self.backend
        exporter_flags = self.exporter_flags
        imapcache = self.imapcache
        imapcache.reopen()
        # get cache importer
        cache_folderid = backend.getCacheFolderId(folderid, imapcache)
        cache_folder_path = imapcache.getFolderPath(cache_folderid)
        if (cache_folder_path == None):
            return None
        importer = imapcache.getContentsImporter(cache_folderid)
        importer.session = backend.session
        importer.store = backend.store
        importer.addressbook = backend.addressbook
    
        state_stream = cache_state.load_state(binascii.hexlify(folderid))

        self.logger.debug("Begin on folder '%s' " % cache_folder_path)
        importer.importer.folder.begin()
    
        exporter = backend.getExporter(folderid)
        try:
            exporter.Config(
                            state_stream, #stream,
                            exporter_flags, importer, None, [PR_ENTRYID, PR_SUBJECT], None, 1)
        except:
            self.logger.debug("exporter config failed on folder '%s' " % cache_folder_path)
            return
        progress = 0
        i = 1
        stream = None
        while True:
            res = exporter.Synchronize(progress);
            if ((i % 5000000000 == 0) or (res[0] == res[1])):
                stream = IStream()
                exporter.UpdateState(stream)
                self.logger.debug("Commit on folder '%s' " % cache_folder_path)
                importer.importer.folder.commit()
                self.logger.debug("Commit completed on folder '%s' " % cache_folder_path)
                cache_state.save_state(binascii.hexlify(folderid), stream)
                imapcache.resync()
                self.logger.debug("ReBegin on folder '%s' " % cache_folder_path)
                importer.importer.folder.begin()
            if (res[0] == res[1]):
                break;
            progress = res[1]
            i += 1
                
        # Make sure to map the syncid to the entryID.
        state = SyncState(stream)
        self.relations[state.syncid] = folderid
        # Make sure the change advisor is monitoring the folder
        try:
            self.advisor.IsMonitoringSyncId(state.syncid)
        except MAPIError:
            self.logger.debug("Adding syncstate '%s' to change advisor" % state)
            self.advisor.AddKeys([state.key])        
        
        imapcache.sync()
        imapcache.close_connection()
        return state

    def update_backend_for_folder_by_folder_path(self, folder_path):
        self.logger.debug("update_backend_for_folder_by_folder_path '%s' get lock" % folder_path)
        self.synclock.acquire()
        self.logger.debug("update_backend_for_folder_by_folder_path '%s' GOT lock" % folder_path)
        exc = True
        try:
            try:
                if self.cache_folder_path_folder_id.has_key(folder_path):
                    folderid = self.cache_folder_path_folder_id[folder_path]
                    self.update_backend_for_folder(folderid)
                else:
                    pass
            except:
                exc = False
                if not self.synclock.release():
                    raise
        finally:
            if exc:
                self.synclock.release()
 
    
    def update_imap_cache_for_folder_by_syncid(self, syncid):
        self.logger.debug("update_imap_cache_for_folder_by_syncid '%s' GETTING lock" % syncid)
        self.synclock.acquire()
        self.logger.debug("update_imap_cache_for_folder_by_syncid '%s' GOT lock" % syncid)
        exc = True
        try:
            try:
                if self.relations.has_key(syncid):
                    folderid = self.relations[syncid]
                    newstate = self.update_imap_cache_for_folder(folderid)
                    if (newstate != None):
                        self.advisesink.UpdateSyncState(newstate)
            except:
                exc = False
                if not self.synclock.release():
                    raise
        finally:
            if exc:
                self.synclock.release()
        self.logger.debug("update_imap_cache_for_folder_by_syncid '%s' RELEASED lock" % syncid)
 
    def update_imap_cache_hierarchy(self):
        cache_state = self.cache_state
        backend = self.backend
        # get backend hierarchy exporter
        exporter = backend.getExporter()
        imapcache = self.imapcache
        importer = imapcache.getHierarchyImporter()
        importer.store = backend.store
        importer.backend = backend
        # get backend exporter
        # get state
        state_stream = cache_state.load_state(backend.root_entryid)
        # config the exporter
        exporter_flags = SYNC_NORMAL | SYNC_UNICODE
        if state_stream == None:  # TODO: check syncstate null
            exporter_flags |= SYNC_NO_SOFT_DELETIONS
        else:
            exporter_flags |= SYNC_READ_STATE
        self.exporter_flags = exporter_flags
        exporter.Config(
                        state_stream, #stream,
                        exporter_flags, importer, None, None, None, 1)
        progress = 0
        while True:
            res = exporter.Synchronize(progress);
            if (res[0] == res[1]):
                break;
            progress = res[1]
        
        if importer.changes > 0:
            stream = IStream()
            exporter.UpdateState(stream)
            streamc = stream.Clone()
            streamc.Seek(0, 0)
            cache_state.save_state(backend.root_entryid, stream)
            imapcache.sync()
            imapcache.reopen()
        
        return importer.changes

    def update_backend_hierarchy(self):
        cache_state = self.cache_state
        backend = self.backend
        # get imapcache hierarchy exporter
        imapcache = self.imapcache
        #imapcache.sync()
        imapcache.reopen()
        exporter = imapcache.getExporter() 

        # get backend importer
        importer = backend.getContentsImporter()
        # get state
        state_stream = cache_state.load_state(backend.root_entryid)
        importer.Config(state_stream, 0)

        syncstate=None
        exporter.Config(importer, syncstate, 0)
        progress = 0
        while True:
            res = exporter.Synchronize(progress);
            if (res[0] == res[1]):
                break;
            progress = res[1]
        
        if exporter.changes > 0:
            stream = IStream()
            importer.UpdateState(stream)
            streamc = stream.Clone()
            streamc.Seek(0, 0)
            cache_state.save_state(backend.root_entryid, stream)
        return exporter.changes

    def update_imap_cache_folders(self):
        backend = self.backend
        folders = backend.getHierarchy()
        for folder in folders:
            self.update_imap_cache_for_folder(folder.id)
        for folder in folders:
            self.update_backend_for_folder(folder.id)

    def check_update_imap_cache_hierarchy(self):
        self.synclock.acquire()
        exc = True
        try:
            try:
                imapcache = self.imapcache
                imapcache.reopen()
                changes = self.update_imap_cache_hierarchy()
                if changes > 0:
                    self.update_imap_cache_folders()
                imapcache.close_connection()
            except:
                exc = False
                if not self.synclock.release():
                    raise
        finally:
            if exc:
                self.synclock.release()
            
    def check_update_backend_hierarchy(self):
        self.synclock.acquire()
        exc = True
        try:
            try:
                imapcache = self.imapcache
                imapcache.reopen()
                changes = self.update_backend_hierarchy()
                if changes > 0:
                    self.update_imap_cache_folders()
                imapcache.close_connection()
            except:
                exc = False
                if not self.synclock.release():
                    raise
        finally:
            if exc:
                self.synclock.release()
            
    def update_imap_cache(self):
        self.logger.debug("update_imap_cache get lock")
        self.synclock.acquire()
        self.logger.debug("update_imap_cache GOT lock")
        exc = True
        try:
            try:
                cache_state = self.cache_state
                
                # hierarchy update ics to imap
                # get backend importer
                zarafa_config = dict(self.config.items('zarafa'))
                zs = ZarafaServer(zarafa_config.get('server_socket', 'https://127.0.0.1:237/zarafa'),
                                  sslkey_file = zarafa_config.get('sslkey_file', '/etc/zarafa/ssl/client.pem'),
                                  sslkey_pass = zarafa_config.get('sslkey_pass', 'secret'))
                backend = zs.get_store(self.backend_account)        # get backend exporter
                self.backend = backend
                # setup change notifications
                stream = None        
                self.advisor = self.backend.getChangeAdvisor()
                self.advisor.Config(stream, None, self.advisesink, 0)
                
                # setup hierarchy change detection
                self.notifsinkmonitor.adviseStore(self, backend.store)
                
                # get cache importer
                imapcache = DsyncStore(None)
                self.imapcache = imapcache
        
                imapcache.connect(self.cache_host, self.cache_ssl_user, self.cache_ssl_password, self.cache_account)
                
                if not cache_state.bind_store(zs, backend, imapcache):
                    self.logger.error("update_imap_cache binding to different server, ignoring account %s" % (self.backend_account))
                    return
        
                self.update_imap_cache_hierarchy()
                
                self.update_backend_hierarchy()
                
                self.update_imap_cache_folders()
                
                imapcache.close_connection()
                self.advisesink.EnableMonitor()
            except:
                traceback.print_exc(file=sys.stdout)
                exc = False
                if not self.synclock.release():
                    self.logger.debug("update_imap_cache RELEASED lock exception")
                    raise
        finally:
            if exc:
                self.synclock.release()
                self.logger.debug("update_imap_cache RELEASED lock")
        
    def stop(self):
        # TODO: stop processing
        pass
        
    
        