# -*- coding: utf-8; indent-tabs-mode: nil -*-

# mailbridge: a mail storage spanning bridge
# 
# Copyright 2011 Ryan Hughes <ryan@n-h.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import MAPI

from MAPI.Util import *
from MAPI.Time import *
from MAPI.Struct import *
import inetmapi
import sys, traceback
import time
import binascii
import logging

from mailbridge.store.Message import Message
from mailbridge.store.Change import ReadState

from mailbridge.util.imapUTF7 import *


class ZarafaStoreExporterProxyContentsImporter(ExchangeImportContentsChanges):
    def __init__(self):
        ExchangeImportContentsChanges.__init__(self, [IID_IExchangeImportContentsChanges])
        self.changes = 0
        self.deletes = 0
        self.readstates = 0

#        self.connection = connection
        self.tagnames = {}
        
        self.importer = None
        self.is_drafts_folder = None
        self.logger = logging.getLogger("ContentsImporter")
        
        for key in MAPI.Tags.__dict__:
            if key.startswith('PR_'):
                try:
                    proptag = MAPI.Tags.__dict__[key]
                    self.tagnames[proptag] = str(key)
                except Exception: pass
                
        for key in MAPICore.__dict__:
            if key.startswith('MAPI_E_') or key.startswith('MAPI_W_'):
                try:
                    err = MAPICore.__dict__[key]
                    self.errornames[err] = str(key)
                except Exception: pass

    def result(self):
        return (self.changes,self.deletes,self.readstates)

    def ImportMessageChange(self,props,flags):
        try:
            self.changes = self.changes + 1
            self.logger.debug("CHANGE %d" % self.changes)
            self.logger.debug("FLAGS %d" % flags)
            entryid = PpropFindProp(props, PR_ENTRYID)
            source_key = PpropFindProp(props, PR_SOURCE_KEY)
            msg = self.store.OpenEntry(entryid.Value, None, MAPI_MODIFY)   # TODO: read only
            deliver_time = msg.GetProps([PR_CREATION_TIME], 0)[0].Value
            message_flags = msg.GetProps([PR_MESSAGE_FLAGS], 0)[0].Value
            flag_status_prop = msg.GetProps([PR_FLAG_STATUS], 0)[0]
            flag_icon_prop = msg.GetProps([PROP_TAG(PT_LONG, 0x1095)], 0)[0]
            icon_index_prop = msg.GetProps([PR_ICON_INDEX], 0)[0]
            sopts = inetmapi.sending_options()
            sopts.no_recipients_workaround = True
            sopts.add_received_date = True
            sopts.headers_only = False
            sopts.force_tnef = False
            sopts.alternate_boundary = None
            sopts.msg_in_msg = False
            
            s=inetmapi.IMToINet(self.session, self.addressbook, msg, sopts)
            msg = Message()
            msg.rfc822 = s
            if (flags == SYNC_NEW_MESSAGE):
                msg.flag_new = 1
            if (message_flags & MSGFLAG_READ == MSGFLAG_READ): msg.flag_read = 1
            else: msg.flag_read = 0
            if (PROP_TYPE(icon_index_prop.ulPropTag)!=MAPICore.PT_ERROR and icon_index_prop.Value == 261): msg.flag_answered = 1
            else: msg.flag_answered = 0
            if (PROP_TYPE(flag_icon_prop.ulPropTag)!=MAPICore.PT_ERROR and flag_icon_prop.Value > 0): msg.flag_flagged = 1
            else: msg.flag_flagged = 0
            if (PROP_TYPE(icon_index_prop.ulPropTag)!=MAPICore.PT_ERROR and icon_index_prop.Value == 262): msg.flag_forwarded = 1
            else: msg.flag_forwarded = 0
            if (self.is_drafts_folder): msg.flag_draft = 1
            msg.received_date = int(deliver_time.unixtime)
            self.importer.importMessageChange(binascii.hexlify(source_key.Value), msg)
        except:
            traceback.print_exc(file=sys.stdout)

        raise MAPIError(SYNC_E_IGNORE)

    def ImportMessageDeletion(self,flags,entries):
        try:
            self.deletes = self.deletes + len(entries)
            hexentries = []
            for entry in entries:
                hexentries.append(binascii.hexlify(entry))
            self.importer.importMessageDeletion(flags, hexentries)
        except:
            traceback.print_exc(file=sys.stdout)

    def ImportPerUserReadStateChange(self,states):
        try:
            readstates = []
            for state in states:
                id = binascii.hexlify(state.SourceKey)
                if (state.ulFlags & MSGFLAG_READ == MSGFLAG_READ): read_flag = 1
                else: read_flag = 0
                readstate = ReadState(id, read_flag)
                readstates.append(readstate)
            self.importer.importMessageReadStateChange(readstates)
        except:
            traceback.print_exc(file=sys.stdout)
        self.readstates = self.readstates + len(states)
        
    def UpdateState(self,stream):
        pass

    def propname(self, proptag):
        try:
            return self.tagnames[proptag]
        except KeyError:
            return '%X' % (proptag)

    def format_value(self, Value):
        if PROP_TYPE(Value.ulPropTag) == PT_BINARY:
            value = bin2hex(Value.Value)
        elif PROP_TYPE(Value.ulPropTag) == PT_ERROR:
            value = self.errorname(Value.Value)
        else:
            value = repr(Value.Value)
            
        return value


class ZarafaStoreExporterProxyHierarchyImporter(ExchangeImportHierarchyChanges):
    def __init__(self, tgtstore):
        ExchangeImportHierarchyChanges.__init__(self, [IID_IExchangeImportHierarchyChanges])
        self.changes = 0
        self.deletes = 0

        self.tgtstore = tgtstore
        self.tagnames = {}
        
        for key in MAPI.Tags.__dict__:
            if key.startswith('PR_'):
                try:
                    proptag = MAPI.Tags.__dict__[key]
                    self.tagnames[proptag] = str(key)
                except Exception: pass
                
        for key in MAPICore.__dict__:
            if key.startswith('MAPI_E_') or key.startswith('MAPI_W_'):
                try:
                    err = MAPICore.__dict__[key]
                    self.errornames[err] = str(key)
                except Exception: pass
    
    def result(self):
        return (self.changes,self.deletes)

    def ImportFolderChange(self, props):
        self.changes = self.changes + 1
        try:
            parent_entryid = PpropFindProp(props, PR_PARENT_ENTRYID)
            entryid = PpropFindProp(props, PR_ENTRYID)
            parent_source_key = PpropFindProp(props, PR_PARENT_SOURCE_KEY)
            source_key = PpropFindProp(props, PR_SOURCE_KEY)
            record_key = PpropFindProp(props, PR_RECORD_KEY)
            display_name = PpropFindProp(props, PR_DISPLAY_NAME)
            container_class = PpropFindProp(props, PR_CONTAINER_CLASS_A)
            hex_record_key = binascii.hexlify(record_key.Value)
            replica_id = self.tgtstore.replica_id
            if (replica_id):
                hex_replica_id = hex(replica_id)[2:]
                guidhex = '0' * (8 - len(hex_replica_id)) + hex_replica_id + '0' * (24 - len(hex_record_key)) + hex_record_key
            else:
                guidhex = '0' * (32 - len(hex_record_key)) + hex_record_key

            folder = self.store.OpenEntry(entryid.Value, None, MAPI_MODIFY)
            dname = folder.GetProps([PR_DISPLAY_NAME], MAPI_UNICODE)[0].Value
            imap_folder_name = None
            set_replica_id = False
            if container_class: 
                s = str(container_class.Value)
                if s == 'IPF.Note':
                    pass
                else:
                    return
            if entryid.Value == self.backend.getDraftsFolderId():
                if self.tgtstore.getDraftsFolderId() == None:
                    imap_folder_name = self.tgtstore.getDraftsFolderName()
                else:
                    self.tgtstore.setFolderPathMapping(self.tgtstore.getDraftsFolderId(), 
                                                             self.tgtstore.getWastebasketFolderName(), 
                                                             binascii.hexlify(source_key.Value))
                    return
            if entryid.Value == self.backend.getReceiveFolderId():
                if self.tgtstore.getReceiveFolderId() == None:
                    imap_folder_name = self.tgtstore.getReceiveFolderName()
                    set_replica_id = True
                else:
                    self.tgtstore.setFolderPathMapping(self.tgtstore.getReceiveFolderId(), 
                                                             self.tgtstore.getWastebasketFolderName(), 
                                                             binascii.hexlify(source_key.Value))
                    return
            if entryid.Value == self.backend.getWastebasketFolderId():
                if self.tgtstore.getWastebasketFolderId() == None:
                    imap_folder_name = self.tgtstore.getWastebasketFolderName()
                else:
                    self.tgtstore.setFolderPathMapping(self.tgtstore.getWastebasketFolderId(), 
                                                             self.tgtstore.getWastebasketFolderName(), 
                                                             binascii.hexlify(source_key.Value))
                    return
            if entryid.Value == self.backend.getSentFolderId():
                if self.tgtstore.getSentFolderId() == None:
                    imap_folder_name = self.tgtstore.getSentFolderName()
                else:
                    self.tgtstore.setFolderPathMapping(self.tgtstore.getSentFolderId(), 
                                                             self.tgtstore.getSentFolderName(), 
                                                             binascii.hexlify(source_key.Value))
                    return

            if type(container_class)==type(None) or str(container_class.Value) == 'IPF.Note':
                parent_folder_guid = self.tgtstore.redis.hget(self.tgtstore.get_redis_basekey()+":sourceid2guid", binascii.hexlify(parent_source_key.Value))
                if imap_folder_name == None:
                    if parent_folder_guid != None:
                        imap_folder_name = self.tgtstore.getFolderPath(parent_folder_guid) + "/" + imapUTF7Encode(display_name.Value.decode('utf-8'))
                    else:
                        imap_folder_name = imapUTF7Encode(display_name.Value.decode('utf-8'))
                folder_guid = self.tgtstore.redis.hget(self.tgtstore.get_redis_basekey()+":sourceid2guid", binascii.hexlify(source_key.Value))
                if folder_guid != None:
                    guidhex = folder_guid
                if (self.tgtstore.getFolderPath(guidhex) != None):
                    # check if this is a moved folder
                    old_imap_folder_name = self.tgtstore.getFolderPath(guidhex)
                    if (old_imap_folder_name != imap_folder_name):
                        # rename folder
                        self.tgtstore.connection.boxrename(guidhex, imap_folder_name, "/")
                        self.tgtstore.setFolderPath(guidhex, imap_folder_name)
                        # rename subfolders
                        paths = self.tgtstore.getFolderPaths()
                        parent_old_imap_folder_name = old_imap_folder_name + '/'
                        for path in paths:
                            if (path[:len(parent_old_imap_folder_name)] == parent_old_imap_folder_name):
                                sub_guidhex = self.tgtstore.getGUIDForFolderPath(path)
                                new_path = imap_folder_name + "/" + path[len(parent_old_imap_folder_name):]
                                if (sub_guidhex):
                                    self.tgtstore.connection.boxrename(sub_guidhex, new_path, "/")
                                    self.tgtstore.setFolderPath(sub_guidhex, new_path)
                    return
                self.tgtstore.connection.boxcreate(imap_folder_name, "/", 0 ,0, guidhex,
                          int(time.time()), 1, 0, 1, 1, "flags" )
                if set_replica_id:
                    self.tgtstore._set_replica_id(guidhex)
                    set_replica_id = False
                self.tgtstore.setFolderPathMapping(guidhex, imap_folder_name, binascii.hexlify(source_key.Value))
        except:
            traceback.print_exc(file=sys.stdout)
        
    def ImportFolderDeletion(self,flags,entries):
        self.deletes = self.deletes + len(entries)
        pass
        
    def UpdateState(self,stream):
        pass

    def propname(self, proptag):
        try:
            return self.tagnames[proptag]
        except KeyError:
            return '%X' % (proptag)

    def format_value(self, Value):
        if PROP_TYPE(Value.ulPropTag) == PT_BINARY:
            value = bin2hex(Value.Value)
        elif PROP_TYPE(Value.ulPropTag) == PT_ERROR:
            value = self.errorname(Value.Value)
        else:
            value = repr(Value.Value)
            
        return value
