# -*- coding: utf-8; indent-tabs-mode: nil -*-

# mailbridge: a mail storage spanning bridge
# 
# Copyright 2011 Ryan Hughes <ryan@n-h.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from mailbridge.store.zarafa.ZarafaServer import ZarafaServer
from mailbridge.store.zarafa.ZarafaStore import ZarafaStore
from mailbridge.store.zarafa.ZarafaStoreMonitor import *
from mailbridge.store.dsync.DsyncStoreMonitor import *
from mailbridge.store.dsync.DsyncStore import DsyncStore


from mailbridge.bridge.Updater import CacheUpdater
from mailbridge.state.RedisStateStorage import RedisStateStreamStorage


import os
import binascii
import time
import logging
import sys
import locale
import threading
import traceback
import ConfigParser


locale.setlocale(locale.LC_ALL, '')


def main(argv = None):
    if argv is None:
        argv = sys.argv

    if len(argv) != 2:
        print >> sys.stderr, 'Usage: <mailbridge-span config-file>'
        return 1
    config = ConfigParser.SafeConfigParser()
    config.read(argv[1])

    dovecot_config = dict(config.items('dovecot', True))
    maillog_pipe_path = dovecot_config.get('maillog_pipe_path')
    
    bridge_config = dict(config.items('bridge', True))
    workers = int(bridge_config.get('workers'))

    logging.basicConfig(level=logging.DEBUG, format='%(asctime)-15s %(levelname)-8s: %(message)s')
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    
    formatter = logging.Formatter('%(asctime)-15s %(levelname)-8s: %(message)s')

    console.setFormatter(formatter)
    
    logging.getLogger('').addHandler(console)
    logging.getLogger('root').addHandler(console)

    sinkmonitor = AdviseSinkMonitor()
    notifsinkmonitor = sinkmonitor.getNotifSinkMonitor()
    maillogmonitor = MaillogMonitor(maillog_pipe_path)

    pool = threadpool.ThreadPool(workers)
    
    for storeinfo in config.items('stores', True):
        zarafa_account = storeinfo[0]
        dovecot_cache = storeinfo[1]
        cache_state = RedisStateStreamStorage()
        cache_updater = CacheUpdater(sinkmonitor, notifsinkmonitor, maillogmonitor, cache_state, logging,
                                     zarafa_account, dovecot_cache, config)
        pool.putRequest(threadpool.WorkRequest(cache_updater.update_imap_cache, None, None))    

    pool.putRequest(threadpool.WorkRequest(sinkmonitor.Run, [pool], None))    
    pool.putRequest(threadpool.WorkRequest(notifsinkmonitor.Run, [pool], None))    
    pool.putRequest(threadpool.WorkRequest(maillogmonitor.Run, [pool], None))    

    while 1: 
        try: 
            time.sleep(0.5)
            pool.poll()
        except KeyboardInterrupt:
            logging.info("User requested exit...")
            cache_updater.stop()
            sinkmonitor.Cancel()
            notifsinkmonitor.Cancel()
            maillogmonitor.Cancel()
        except threadpool.NoResultsPending: 
            break

    logging.debug("Threadpool ready, waiting for workers to exit")
    pool.dismissWorkers(workers, True)
    logging.debug("Done")


if __name__ == '__main__':
    sys.exit(main())
